;; Generated package description from blist.el  -*- no-byte-compile: t -*-
(define-package "blist" "0.3" "Display bookmarks in an ibuffer way" '((ilist "0.1") (emacs "24")) :commit "9e1054d3cb2b796704ebeb2575b29ad70f0bb21c" :authors '(("Durand" . "durand@jsdurand.xyz")) :maintainer '("Durand" . "durand@jsdurand.xyz") :keywords '("convenience") :url "https://gitlab.com/mmemmew/blist")
